(* MIDI-to-SX-150 Interface.
   Thomas Henry -- 02/25/2015
   Version 1.0

   This module responds to MIDI note numbers 36 through 84,
   corresponding to C2 on up to C6. Numbers larger or smaller
   than these extremes are constrained to the limits.

   Usable commands are:

      Note On
      Note Off
      Omni On
      Omni Off
      All Notes Off

   All other commands are ignored, but I've left program
   stubs in place for them should you think of some way
   of using them.
*)

program Gakken_Interface;

{$PROCESSOR PIC16F88}
{$OPTIMIZE SPEED}
{$FREQUENCY 20 MHZ}
{$CONFIG BOREN = OFF, CCPMX = RB0, CP = OFF, CPD = OFF,
         FOSC = HS, LVP = OFF, MCLRE = ON, PWRTE = OFF,
         WDTE = OFF, WRT = OFF}       // CONFIG1
{$CONFIG FCMEN = OFF, IESO = OFF}     // CONFIG2
{$INTERRUPTS UNIQUE}                  // only MIDI interrupts appear
{$INIT ANALOGS}

uses
  Serial;                             // MIDI is 31.250 kHz serial

const
  NOTE_OFF          = $80;            // the MIDI command status bytes
  NOTE_ON           = $90;
  POLY_KEY_PRESSURE = $A0;
  CTRL_CHANGE       = $B0;
  PGM_CHANGE        = $C0;
  CHANNEL_PRESSURE  = $D0;
  PITCH_WHEEL       = $E0;
  SYS_EX            = $F0;

  ALL_NOTES_OFF     = $7B;            // controller data byte
  OMNI_OFF          = $7C;
  OMNI_ON           = $7D;

  Q_SIZE = 80;                        // size of queue in bytes
  OFFSET = 10;                        // Gakken control voltage offset

var
  red : boolean @ PortA.RA2;          // red LED element
  green : boolean @ PortA.RA3;        // green LED element
  select : boolean @ PortA.RA4;       // select pushbutton switch

  runStat, seen1 : boolean;
  current, head, tail, midiCmd, midiChn,
  note, bottom, top, mid, total : byte;
  channel : byte;                     // $FF signifies omni mode
  channelSave : eeprom byte = 0;      // default channel 1 (zero) at burn
  queue : array[Q_SIZE] of byte;

//----- Procedures

procedure fetch; interrupt;
// The queue is constantly filled in the background by the
// interrupt procedure. It is emptied in the main program.

  var
    temp : byte;

  begin
    temp := SerialPort_Input;               // read a byte in
    queue[tail] := temp;                    // stuff into queue
    inc(tail);                              // prep for next time
    if (tail > (Q_SIZE-1)) then             // wrap around if needed
      clr(tail)
  end;

//----- Main Program

begin
  PORTA := 0;                               // clear outputs to start
  PORTB := 0;
  TRISA := 0b10110000;                      // and set the ports
  TRISB := 0b00000100;

  baud(31250, RX);                          // receive only
  clr(head);                                // initialize queue
  clr(tail);
  PIE1 := [RCIE];                           // enable serial RX interrupts
  INTCON := [GIE, PEIE];                    // enable at highest level, too

  //----- Selection Business At Reset

  if (not select) then                      // if button is pushed
    begin
      channel := $FF;                       // begin from count before zero
      repeat                                // repeat while button pushed
        inc(channel);                       // bump channel number
        red := 0;                           // LEDs off
        green := 0;
        delay_ms(500);                      // for a half-second

        if (channel < 16) then              // count the channel numbers
          red := 1                          // red means channel
        else
          green := 1;                       // green means omni

        delay_ms(500);                      // wait a half-second

        if (channel = 16) then              // omni signified by $FF
          channel := $FF;
      until (select);                       // keep going until not pushed

      channelSave := channel;               // save in eeprom for next time
      delay_ms(10);                         // let the write settle
    end
  else                                      // button not pushed
    begin                                   // so, display the channel
      channel := channelSave;               // recover channel from eeprom
      delay_ms(10);
      if (channel = $FF) then               // indicate omni on
        begin
          red := 0;                         // green is on
          green := 1
        end
      else                                  // else, a particular channel
        begin
          green := 0;
          for current := 1 to channel do    // flash the channel
            begin
              red := 1;                     // blink red LED the number
              delay_ms(500);                // of the channel
              red := 0;
              delay_ms(500)
            end;
            red := 1;                       // conclude with LED on
        end
    end;

  //----- MIDI Processing Loop

  loop
    if (head <> tail) then                  // anything in queue?
      begin
         current := queue[head];            // if so, get current byte
        inc(head);                          // bump for next time
        if (head > Q_SIZE-1) then           // or wrap around to start
          clr(head);

        //----- handle status byte

        if (current.7) then                 // if a status byte
          if (current >= $F0) then          // if a System message
            if (current >= $F8) then        // if System Realtime message
              goto done                     // quick exit
            else
              runStat := false              // System Common or Exclusive
          else                              // else a Channel message
            begin
              midiCmd := current and $F0;   // strip the command
              midiChn := current and $0F;   // strip the channel
              runStat := true;              // assume running status on
              // allow if in omni mode, or the channel matches
              if (not((channel = $FF) or (midiChn = channel))) then
                runStat := false;           // it's not, so running status off
              seen1 := false                // haven't seen first datum yet
            end

        //----- handle data bytes

        else                                // else it's a data byte
          if (runStat) then                 // proceed if running status on
            begin
              CASE midiCmd of               // one data byte commands
                PGM_CHANGE :                // STUB: Program Change
                  begin
                    goto done               // quick exit
                  end;

                CHANNEL_PRESSURE :          // STUB: Channel Pressure
                  begin
                    goto done               // quick exit
                  end
              end; {case}

              if (seen1) then               // got both bytes now
                begin
                  seen1 := false;           // get set for next time

                  case midiCmd of           // two data byte commands
                    NOTE_ON :
                      begin
                        if (current) then   // if a non-zero velocity
                          begin
                            if (note < 36) then // stay within limits
                              note := 36
                            elsif (note > 84) then
                              note := 84;

                            // translate to Gakken range
                            note := note - 36 + OFFSET;

                            bottom := note and 0b00000011;
                            mid := (note shl 1) and 0b00011000;
                            top := (note shl 2) and 0b11000000;
                            total := bottom or mid or top;
                            PortB := total; // send to DAC
                            PortA.RA0 := 1;
                          end
                        else                // else, zero velocity
                          begin
                            PortB := 0;     // clear the DAC
                            PortA.RA0 := 0
                          end
                      end;

                    NOTE_OFF :
                      begin
                        PortB := 0;         // clear the DAC
                        PortA.RA0 := 0
                      end;

                    CTRL_CHANGE :
                      begin
                        if (note = OMNI_ON) then
                          begin
                            channel := 0xFF;
                            red := 0;       // indicate omni on
                            green := 1
                          end
                        elsif (note = OMNI_OFF) then
                          begin
                            channel := 0;   // default to channel 1
                            red := 1;       // indicate omni off
                            green := 0
                          end
                        elsif (note = ALL_NOTES_OFF) then
                          begin
                            PortB := 0;     // clear the DAC
                            PortA.RA0 := 0
                          end
                      end;

                    PITCH_WHEEL :           // all else are stubs
                      begin                 // and ignored
                        nop
                      end;

                    POLY_KEY_PRESSURE :
                      begin
                        nop
                      end;
                  end {case}
                end {got both bytes}
              else                          // else need one more datum
                begin
                  note := current;          // save first datum
                  seen1 := true             // flag that we've got it
                end;
              done:
            end {running status on}
      end {if anything in queue}
  end {loop}											          // repeat in perpetuity
end. {program}